    cd h:\abm
    set seed 1234567
    set rmsg on
    clear all
    // -----------------------------------class definitions
    // import abm_grid class definition
    do abm_grid.mata
    
    local vulnerable = 0
    local infected   = 1
    local infectious = 2
    local healed     = 3
    
    // this allows you to change or add to the grid class when needed
    mata
    class grid extends abm_grid
    {
    
    }
    
    // the main agent based model
    class abm
    {
        private:
            class grid scalar universe
            class AssociativeArray scalar agents

            real scalar  not_vaccinated
            real scalar  tdim
            real scalar  radius
            real matrix  infected
            real scalar  n_infected
            
            
            void         setup_grid()
            void         setup_agents()
            void         create_agents()
            void         populate_universe()
            void         copy_agents()
            void         infect()
            real scalar  status()
            real scalar  is_vaccinated()
            void         initial_infection()
            
        public:
            transmorphic rdim()
            transmorphic cdim()
            transmorphic tdim()
            transmorphic torus()
            transmorphic not_vaccinated()
            transmorphic radius()
            void         run()
            real scalar  n_infected()
            void         extract()
    }

    // --------------------------------------setup the grid
    transmorphic abm::rdim(| real scalar dim) 
    {
        if (args() == 1) {
            universe.rdim(dim)
        }
        else {
            return(universe.rdim())
        }
    }
    transmorphic abm::cdim(| real scalar dim) 
    {
        if (args() == 1) {
            universe.cdim(dim)
        }
        else {
            return(universe.cdim())
        }
    }
    transmorphic abm::torus(| real scalar val)
    {
        if (args() == 1) {
            universe.torus(val)
        }
        else {
            return(universe.torus())
        }
    }
    void abm::setup_grid()
    {
        if (universe.torus() == . ) {
            universe.torus(1)
        }
        universe.setup()
    }
    
    // ------------------------------------- setup agents
    
    transmorphic abm::not_vaccinated(| real scalar dim) 
    {
        if (args() == 1) {
            if (dim <= 0 | mod(dim,1) != 0 ) {
                _error(3000, "dim must be an integer larger than 0")
            }
            
            not_vaccinated = dim
        }
        else {
            return(not_vaccinated)
        }
    }    
    transmorphic abm::tdim(| real scalar dim)
    {
        if (args() == 1) {
            if (dim <= 0 | mod(dim,1) != 0) {
                _error(3000, "dim must be an integer larger than 0")
            }
            tdim = dim
        }
        else {
            return(tdim)
        }
    }
    
    transmorphic abm::radius(| real scalar number)
    {
        if (args() == 1) {
            if (number <= 0 | mod(number,1) != 0) {
                _error(3000, "number must be an integer larger than 0")
            }
            radius = number
        }
        else {
            return(radius)
        }
    }

    void abm::setup_agents()
    {
        if (tdim == .) {
            _error(3000, "tdim must be specified")
        }
        if (not_vaccinated == .) {
            _error(3000, "not_vaccinated must be specified")
        }
        if (not_vaccinated > universe.size()) {
            _error(3000, "specified more not vaccinated agents than cells on grid")
        }
        if (radius == .) {
            radius = 2
        }
        
        agents.reinit("real", 2)
    }
    
    // -------------------------------- helper functions
    
    real scalar abm::status(real scalar r, real scalar c, real scalar t)
    {
        real scalar    id
        real rowvector key
        
        if (universe.has_agent(r,c)) {
            id = universe.agent_id(r,c)
            key = id, t
            return(agents.get(key))    
        }
        else {
            return(0)
        }
    }
    
    real scalar abm::is_vaccinated(real scalar r, real scalar c, real scalar t)
    {
        return(!universe.has_agent(r,c))
    }

    void abm::copy_agents(real scalar t) 
    {
        real scalar    i, content
        real rowvector key
        
        for(i=1 ; i <= not_vaccinated ; i++) {
            key = i,t
            content = agents.get(key)
            key = i,t+1
            agents.put(key,content)
        }
    }
    
    real scalar abm::n_infected() {
        return(n_infected)
    }
    
    void abm::extract(string rowvector names) 
    {
        real matrix grid, ag
        real rowvector idx
        real colvector status
        real scalar i
        
        if (cols(names) != 5) {
            _error(3000, "5 names need to be specified")
        }
        grid = universe.extract()
        _sort(grid,5)
        grid = grid[.,(1,2)]
        grid = J(tdim+1,1,grid)
        
        ag = agents.keys()
        _sort(ag,(2,1))
        
        status = J(rows(ag),1,.)
        for(i=1 ; i<= rows(ag); i++) {
            status[i] = agents.get(ag[i,.])
        }
        ag = ag, status, grid
        
        idx = st_addvar("float", names)

        if (st_nobs() < rows(ag) ) {
            st_addobs(rows(ag) - st_nobs())
        }
        st_store(.,idx,ag)
    }
    
    // ----------------------------------- run the model
    
    void abm::create_agents() 
    {
        real scalar i, id
        real matrix unvac
        
        for (i = 1 ; i <= not_vaccinated ; i++) {
            agents.put((i,1),0)
        }
    }
    
    void abm::populate_universe()
    {
        real scalar    i
        real matrix grid
        
        grid = jumble(universe.schedule())

        infected = grid[1,.]
        
        for(i = 1 ; i <= not_vaccinated ; i++) {
            universe.create_agent(grid[i,1],grid[i,2],i)
        }
    }
    
    void abm::initial_infection()
    {
        real scalar id

        id = universe.agent_id(infected[1], infected[2])
        agents.put((id,1),`infected')
        n_infected = 1
    }
    
    void abm::infect(real scalar t)
    {
        real scalar    i, id, r, c, j, rn, cn, idn
        real rowvector key
        real colvector tokeep
        real matrix    neighbours, toadd

        tokeep = J(rows(infected),1,1)
        toadd = J(0,2,.)
        for(i = 1 ; i <= rows(infected) ; i++) {
            r = infected[i,1]
            c = infected[i,2]
            id = universe.agent_id(r, c)
            
            if ( status(r, c, t) == `infectious' ) {
                agents.put((id,t+1),`healed')
                tokeep[i] = 0
            }
            else{
                key = id, t+1
                agents.put(key, `infectious')
                neighbours = universe.find_spiral(r,c,radius)
                
                for (j = 1 ; j <= rows(neighbours); j++) {
                    rn = neighbours[j,1]
                    cn = neighbours[j,2]
                    if (  is_vaccinated(rn,cn,t+1) == 0 & status(rn,cn,t+1) == 0  ) {
                        idn = universe.agent_id(rn,cn)
                        key = idn,t + 1
                        agents.put(key,`infected')
                        n_infected = n_infected + 1
                        toadd = toadd \ (rn,cn)
                    }
                }
            }
        }
        infected = select(infected, tokeep)
        infected = infected \ toadd
    }
    
    void abm::run()
    {    
        real scalar t

        setup_grid()
        setup_agents()
        create_agents()
        populate_universe()
        initial_infection()
        for(t=1; t <= tdim ; t++) {
            copy_agents(t)
            infect(t)
        }
    }
    end
    
    // run a single model and graph the evolution of the infection over time
    mata
    foo = abm()
    foo.rdim(40)
    foo.cdim(40)
    foo.tdim(20)
    foo.radius(2)
    foo.not_vaccinated(400)
    foo.run()
    foo.extract(("id", "time", "status", "y", "x"))
    end
    
    replace y = - y

    mata st_local("rdim", strofreal(foo.rdim()))
    mata st_local("cdim", strofreal(foo.cdim()))
    mata st_local("tdim", strofreal(foo.tdim()))
    tempfile data
    save `data'
    drop _all
    set obs `rdim'
    gen x = _n
    expand `cdim'
    bys x : gen y = -_n
    expand `tdim' + 1
    bys y x : gen time = _n
    merge 1:1 x y time using `data'
    assert _merge != 2
    drop _merge
    
    forvalues t = 1/21 {
        scatter y x if status == 0 & time == `t' ,    ///
            msymbol(Oh)                            || ///
        scatter y x if status == 1 & time == `t' ,    ///
            msymbol(S) mcolor(gs8)                 || ///
        scatter y x if status == 2 & time == `t' ,    ///
            msymbol(S)                             || ///
        scatter y x if status == 3 & time == `t' ,    ///
            msymbol(+)                             || ///
        scatter y x if status == . & time == `t',     ///
            msymbol(o) msize(*.10) mcolor(gs8)        ///
            aspect(1) name(gr`t', replace)            ///
            yscale(off range(-41.5 -0.5))             ///
            xscale(off range(.5 41.5))                ///
            plotregion(margin(zero))                  ///
            legend(order( 1 "vulnerable"              ///
                          2 "infected"                ///
                             3 "infectious"              ///
                          4 "healed/immunized"        ///
                          5 "immunized"))
    }
    
    // run multiple models and plot the seriousness of the infection against 
    // the immunization rate
    mata
    scen = (16, 80, 120, 160, 200, 240, 280, 320, 360, 400)
    res = J(200,2,.)
    k=0
    for (i = 1; i <= 10 ; i++) {
        printf("number of infected: %3.0f \n", scen[i])
        foo.not_vaccinated(scen[i])
        for(j = 1; j<= 20; j++) {
            k = k +1
            foo.run()
            res[k,.] = scen[i], foo.n_infected()
        }
    }
    
    idx =  st_addvar("float", ("N","n"))
    st_addobs(200)
    st_store((1..200)',idx,res)
    end
    
    gen perc_notvac = N/(40*40)*100
    gen perc_inf    = n/N*100
    
    scatter perc_inf perc_notvac, msymbol(o) mcolor(%20) ///
        xtitle("percentage not vaccinated")              ///
        ytitle("percentage infected of not vacinated ")
        
    