clear all

do abm_grid.mata

mata
mata set matastrict on

class grid extends abm_grid
{

}

class schelling {
	private:
		class grid             scalar universe
		class AssociativeArray scalar agents

		real   scalar    rdim
		real   scalar    cdim
		real   scalar    tdim
		real   scalar    limit
		real   scalar    nblue
		real   scalar    nred
		real   scalar    nagents
		
		void             is_posint()
		void             is_prop()
		
		void             setup_agents()
		void             setup_universe()		
		void             populate_universe()
		
		string scalar    get_color()
		real   scalar    sum_neighbours()	
		real   scalar    is_happy()
		void             move()		
		void             step()	
		

	
	public:
		transmorphic     rdim()
		transmorphic     cdim()
		transmorphic     tdim()
		transmorphic     limit()
		transmorphic     nblue()
		transmorphic     nred()
		
		void             run()
		
		void             extract()
		real matrix      summ()
		void             summtable()
}


// -------------------------------------------------------------- error messages
void schelling::is_posint(real scalar val, string scalar name) 
{
	string scalar errmsg
	
	if (val <= 0 | mod(val,1)!= 0 ){
		errmsg = name + " must be a positive integer"
		_error(3300, errmsg)
	}
}
void schelling::is_prop(real scalar val, string scalar name) 
{
	string scalar errmsg
	
	if (val < 0 | val > 1 ){
		errmsg = name + " must be between or including 0 and 1"
		_error(3300, errmsg)
	}
}

// ------------------------------------------------------------------ setup grid
transmorphic schelling::rdim(| real scalar dim) 
{
	if (args() == 1) {
		is_posint(dim,"dim")
		universe.rdim(dim)
		rdim = dim
	}
	else {
		return(universe.rdim())
	}
}
transmorphic schelling::cdim(| real scalar dim) 
{
	if (args() == 1) {
		is_posint(dim,"dim")
		universe.cdim(dim)
		cdim = dim
	}
	else {
		return(universe.cdim())
	}
}
transmorphic schelling::tdim(| real scalar dim) 
{
	if (args() == 1) {
		is_posint(dim,"dim")
		universe.tdim(dim)
		tdim = dim
	}
	else {
		return(universe.tdim())
	}
}
void schelling::setup_universe()
{
	if (tdim()==.) {
		_error(3000, "tdim must be set first")
	}
	universe.setup()
}	

// ---------------------------------------------------------------- setup agents
transmorphic schelling::limit(| real scalar val) 
{
	if (args() == 1) {
		is_prop(val,"value")
		limit = val
	}
	else {
		return(limit)
	}
}
transmorphic schelling::nblue(| real scalar val) 
{
	if (args() == 1) {
		is_posint(val,"value")
		nblue = val
	}
	else {
		return(nblue)
	}
}
transmorphic schelling::nred(| real scalar val) 
{
	if (args() == 1) {
		is_posint(val,"value")
		nred = val
	}
	else {
		return(nred)
	}
}
void schelling::setup_agents() 
{
	real   scalar k, i
	string scalar errmsg
	
	if (limit == .) {
		_error(3300, "limit has not been set")
	}
	if (nblue == .) {
		_error(3300, "number of blue agents has not been set")
	}
	if (nred == .) {
		_error(3300, "number of red agents has not been set")
	}
	nagents = nred + nblue
	if (nagents > universe.size()) {
		errmsg = "you specified " + strofreal(nagents) + " agents on a " +
		         strofreal(rdim()) + " by " + strofreal(cdim()) + " grid"
		_error(3300, errmsg)
	}

	agents.reinit("real")
	
	k = 0
	for (i = 1 ; i <= nred ; i++) {
		k = k +1
		agents.put(k, "red")
	}
	for (i = 1 ; i <= nblue ; i++) {
		k = k +1
		agents.put(k, "blue")
	}
}

// --------------------------------------------------------------- run the model
void schelling::populate_universe(){
	real matrix grid0
	real scalar i, r, c
	
	grid0 = universe.schedule()
	_jumble(grid0)
	for(i = 1 ; i <= nagents ; i++) {
		r = grid0[i,1]
		c = grid0[i,2]

		universe.create_agent(r,c,i,0)
	}
}


void schelling::step(real scalar t, real scalar r, real scalar c) 
{
	if (is_happy(t,r,c,t,r,c)==0) {
		move(t,r,c)
	}
}

void schelling::run() {
	real scalar  t, k, r, c
	real matrix coords
	
	setup_universe()
	setup_agents()
	populate_universe()
	
	for(t=0; t < universe.tdim() ; t++) {
		universe.copy_grid(t, t+1)
		coords = universe.schedule()
		for(k=1; k <= rows(coords); k++){
			r = coords[k,1]
			c = coords[k,2]
			if (universe.has_agent(r,c,t)) {
				step(t, r, c)
			}
		}
	}
}

// ------------------------------------------------------------ helper functions

string scalar schelling::get_color(real scalar t, real scalar r, real scalar c)
{
	real scalar    id
	
	if (universe.has_agent(r,c,t)) {
		id = universe.agent_id(r, c,t)
	}
	else {
		_error(3300, "no agent in this cell")
	}
	return(agents.get(id))
}

real scalar schelling::is_happy(
	real scalar t , real scalar r , real scalar c , 
	real scalar ts, real scalar rs, real scalar cs)
{
	real   scalar prop
		
	prop   = sum_neighbours(t,r,c,ts,rs,cs)
	
	return(prop >= limit)
}

real scalar schelling::sum_neighbours(
                                real scalar t, real scalar r, real scalar c, 
								real scalar ts, real scalar rs, real scalar cs) 
{
	real   scalar    k, s, i, j, l, isself
	real   matrix    neigh
	string scalar    self
	
	self = get_color(ts,rs,cs)
	
	neigh = universe.find_ring(r,c,1)
	k = 0
	s = 0
	for (l=1 ; l <=rows(neigh) ; l++) {
		i = neigh[l,1]
		j = neigh[l,2]
		isself = (i==rs & j==cs)
		if (universe.has_agent(i,j,t) & !isself) {
			k = k + 1
			s = s + (get_color(t,i,j)==self)
		}
	}
	return((k==0 ? 1 : s/k))
}

void schelling::move(real scalar t, real scalar r, real scalar c)
{
	real scalar    radius, i, stop, rd, cd, maxradius
	real matrix    neigh
	
	stop = 0
	maxradius = max( ( rdim-r, r-1, cdim-c, c-1 ) )

	for (radius = 1 ; radius <= maxradius ; radius ++) {
		neigh = universe.find_ring(r,c,radius)
		for (i = 1; i <= rows(neigh); i++) {
			rd = neigh[i,1]
			cd = neigh[i,2]
			if (universe.has_agent(rd,cd, t+1,) == 0) {
				if( is_happy(t+1,rd, cd, t, r, c) ) {
					universe.move_agent(r,c, rd, cd, t+1)
					stop = 1
					break
				}
			}
		}
		if (stop == 1) break
	}
}

// ---------------------------------------------------------------------- export
void schelling::extract(string rowvector names)
{
	real matrix res
	real colvector color, happy
	real rowvector idx
	real scalar i, k, t,r,c
	
	if(cols(names)!=6){
		_error(3300, "names must contain 6 columns")
	}
	
	res = universe.extract()
	color = J(rows(res),1,.)
	happy = J(rows(res),1,.)
	k = cols(res)
	for (i = 1 ; i <= rows(res); i++) {
		color[i] = agents.get(res[i,k]) == "red"
		t = res[i,3]
		r = res[i,1]
		c = res[i,2]
		happy[i] = is_happy(t,r,c,t,r,c)
	}
	res = res[.,1..3], res[.,5], color, happy
	
	idx = st_addvar("float", names)

	if (st_nobs() < rows(res) ) {
		st_addobs(rows(res) - st_nobs())
	}
	st_store(.,idx,res)
}

real matrix schelling::summ()
{
	real matrix res, prop
	real scalar t, r, c, k
	res = J(tdim+1,3,.)
	res[.,1] = (0 .. tdim)'

	for(t=0 ; t<= tdim; t++) {
		prop = J(nagents, 2, .)
		k = 1
		for(r=1; r <= rdim; r++) {
			for (c=1 ; c<=cdim ; c++ ) {
				if (universe.has_agent(r,c,t)) {
					prop[k,1] = sum_neighbours(t,r,c, t, r, c)
					prop[k,2] = is_happy(t,r,c,t,r,c)
					k = k + 1
				}
			}
		}
		res[|t+1,2 \ t+1,3|] = mean(prop)
	}
	return(res)
}


void schelling::summtable(| string scalar matname)
{
	real matrix res
	string scalar cmd
	
	res = summ()
	if (args()==0) {
		matname = st_tempname()
	}
	st_matrix(matname,res)
	st_matrixcolstripe(matname, (J(3,1, "") , ("Iteration" \ "prop_same" \ "prop_happy")) )
	cmd = "matlist " + matname + ", underscore format(%10.3g) names(columns) " +
          "border(rows) " + 
		  "title(Average proportion of neighbours with same color and average proportion of happy agents)"
	stata(cmd)
}


end

